package avatartranslator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class AvatarDataType {
    public final static int INTEGER = 0;
    public final static int BOOLEAN = 1;
    public final static int DATATYPE = 2;
    public static Map<String,AvatarDataType> dataTypeMap = new HashMap<String,AvatarDataType>();
    private String name;
    private List<String> fieldNames = new ArrayList<String>();
    private List<Integer> fieldTypes = new ArrayList<Integer>();
    private List<String> dataTypeNames = new ArrayList<String>();
    private int size = -1;

    public AvatarDataType(String _name) {
        name = _name;
    }
    public int getFullSize() { return size; }
    public int getNbFields() { return fieldNames.size(); }
    String getFieldName(int _i) { return fieldNames.get(_i); }
    public int getFieldType(int _i) { return fieldTypes.get(_i); }
    String getDataTypeNameName(int _i) { return dataTypeNames.get(_i); }
    public void addField(String _fieldName, int _fieldType, String _dataTypeName) {
        if (size != -1) return; // DataType has already been finalized
        if (_fieldType == AvatarDataType.INTEGER || _fieldType == AvatarDataType.BOOLEAN)
            dataTypeNames.add(null);
        else if (_fieldType != DATATYPE)
            return;
        else
            dataTypeNames.add(_dataTypeName);
        fieldTypes.add(_fieldType);
        fieldNames.add(_fieldName);
    }
    public static void finalize(AvatarDataType adt) {
        if (adt.size != -1) return;
        int size = 0;
        int length = adt.getNbFields();
        for (int i = 0; i < length; i++) {
            if (adt.getFieldType(i) == AvatarDataType.INTEGER || adt.getFieldType(i) == AvatarDataType.BOOLEAN) {
                size += 1;
                continue;
            }
            AvatarDataType fieldDataType = dataTypeMap.get(i);
            if (fieldDataType == null)  return;
            finalize(fieldDataType); // loops if recursion => recursive types are forbidden
            int fieldSize = fieldDataType.getFullSize();
            if (fieldSize == -1)  return;
            size += fieldSize;
        }
        adt.size = size;
    }
}
