/* Copyright or (C) or Copr. GET / ENST, Telecom-Paris, Ludovic Apvrille
 * 
 * ludovic.apvrille AT enst.fr
 * 
 * This software is a computer program whose purpose is to allow the
 * edition of TURTLE analysis, design and deployment diagrams, to
 * allow the generation of RT-LOTOS or Java code from this diagram,
 * and at last to allow the analysis of formal validation traces
 * obtained from external tools, e.g. RTL from LAAS-CNRS and CADP
 * from INRIA Rhone-Alpes.
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */




package ui.atd;

//import java.awt.*;

import attacktrees.Attack;
import attacktrees.AttackElement;
import attacktrees.AttackNode;
import attacktrees.AttackTree;
import avatartranslator.AvatarElement;
import avatartranslator.avatarucd.AvatarActor;
import myutil.NameChecker;
import myutil.TraceManager;
import ui.*;
import ui.ucd.TGConnectorInclude;
import ui.ucd.TGConnectorUseCase;
import ui.ucd.UCDActor;

import java.awt.*;
import java.util.*;

//import org.w3c.dom.*;
//import org.xml.sax.*;
//import javax.xml.parsers.*;

/**
 * Class AttackTreePanel
 * Panel used for drawing attack trees
 * Creation: 03/11/2009
 * @version 1.0 03/11/2009
 * @author Ludovic APVRILLE
 */
public class AttackTreeDiagramPanel extends TDiagramPanel implements TDPWithAttributes, NameChecker.SystemWithNamedElements {
    
    public  AttackTreeDiagramPanel(MainGUI mgui, TToolBar _ttb) {
        super(mgui, _ttb);
        //addComponent(400, 50, TGComponentManager.EBRDD_START_STATE, false);
        /*TDiagramMouseManager tdmm = new TDiagramMouseManager(this);
        addMouseListener(tdmm);
        addMouseMotionListener(tdmm);*/
    }
    
    public boolean actionOnDoubleClick(TGComponent tgc) {
        return true;
    }
    
    public boolean actionOnAdd(TGComponent tgc) {
        return false;
    }
    public boolean actionOnValueChanged(TGComponent tgc) {
        return false;
    }
    
    public  boolean actionOnRemove(TGComponent tgc) {
        return false;
    }
    
    public String getXMLHead() {
        return "<AttackTreeDiagramPanel name=\"" + name + "\"" + sizeParam() + zoomParam() + " >";
    }
    
    public String getXMLTail() {
        return "</AttackTreeDiagramPanel>";
    }
    
    public String getXMLSelectedHead() {
        return "<AttackTreeDiagramPanelCopy name=\"" + name + "\" xSel=\"" + xSel + "\" ySel=\"" + ySel + "\" widthSel=\"" +
                widthSel + "\" heightSel=\"" + heightSel +
                "\"" + zoomParam()  + ">";
    }
    
    public String getXMLSelectedTail() {
        return "</AttackTreeDiagramPanelCopy>";
    }
    
    public String getXMLCloneHead() {
        return "<AttackTreeDiagramPanelCopy name=\"" + name + "\" xSel=\"" + 0 + "\" ySel=\"" + 0 + "\" widthSel=\""
                + 0 + "\" heightSel=\"" + 0 +
                "\"" + zoomParam()  + ">";
    }
    
    public String getXMLCloneTail() {
        return "</AttackTreeDiagramPanelCopy>";
    }
    
    public void makeGraphicalOptimizations() {
        // Segments of connector that mask components
        
        // Components over others
        
        // Position correctly guards of choice
    }
    
    public LinkedList<TGComponent> getAllAttacks() {
		LinkedList<TGComponent> list = new LinkedList<TGComponent>();
		TGComponent tgc;
		
		ListIterator iterator = getComponentList().listIterator();
		
		while(iterator.hasNext()) {
            tgc = (TGComponent)(iterator.next());
            if (tgc instanceof ATDAttack) {
				list.add(tgc);
			}
		}
		
		return list;
		
	}
	
	public boolean hasAutoConnect() {
		return false;
	}
	
	public void setConnectorsToFront() {
		TGComponent tgc;
		
		//
		
        Iterator iterator = componentList.listIterator();
        
		ArrayList<TGComponent> list = new ArrayList<TGComponent>();
		
        while(iterator.hasNext()) {
            tgc = (TGComponent)(iterator.next());
			if (!(tgc instanceof TGConnector)) {
				list.add(tgc);
			}
		}
		
		//
		for(TGComponent tgc1: list) {
			//
			componentList.remove(tgc1);
			componentList.add(tgc1);
		}
	}

    public String drawFromAttackTreeModel(AttackTree _at)  {
        Attack root = null;

        // find the root attack
        for(Attack att: _at.getAttacks()) {
            if (att.isRoot()) {
                root = att;
                break;
            }
        }

        if (root == null) {
            return "No root attack";
        }

        HashMap<AttackElement, TGComponent> mapOfComponents = new HashMap<>();


        return drawAttackFromAttackTreeModel(mapOfComponents, _at, root, null, -1, 600, 100, 1200);
    }

    public String drawAttackFromAttackTreeModel(HashMap<AttackElement, TGComponent> _mapOfComponents, AttackTree _at, Attack _att,
                                                ATDConstraint _const, int _valueConst, int _x, int _y, int _length) {

        // We draw the attack
        ATDAttack attack = new ATDAttack(_x, _y, getMinX(), getMaxX(), getMinY(), getMaxY(), true, null, this);
        attack.setValue(_att.getName());
        attack.setRootAttack(_att.isRoot());
        attack.setDescription(_att.getDescription());
        addBuiltComponent(attack);
        _mapOfComponents.put(_att, attack);

        // If the constraint is non-null, we link the attack to the constraint
        if (_const != null) {
            TGConnectingPoint p1 = attack.closerFreeTGConnectingPoint(_const.getX(), _const.getY(), false, true);
            TGConnectingPoint p2 = _const.closerFreeTGConnectingPoint(attack.getX(), attack.getY(), true, false);

            if (p1 != null && p2 != null) {
                p1.setFree(false);
                p2.setFree(false);

                TGConnector conn = null;
                Vector<Point> points = new Vector<Point>();
                conn = new ATDAttackConnector(0, 0, 0, 0, 0, 0, true, null,
                        this, p1, p2, points);
                if (_valueConst != -1) {
                    conn.setValue("" + _valueConst);
                }

                if (conn != null) {
                    addComponent(conn, 0, 0, false, true);
                }
            }
        }

        // We look for the constraint linked to this attack
        AttackNode nextNode = null;
        for(AttackNode node: _at.getAttackNodes()) {
            if (node.getResultingAttack() == _att) {
                nextNode = node;
                break;
            }
        }

        // Leaf attack?
        if (nextNode == null) {
            return null;
        }

        // We create the ATDConstraint
        _y = _y + 70;
        ATDConstraint constraint = new ATDConstraint(_x, _y, getMinX(), getMaxX(), getMinY(), getMaxY(), true, null, this);
        //constraint.setValue(nextNode.getName());
        constraint.setType(nextNode.getType());
        addBuiltComponent(constraint);
        _mapOfComponents.put(nextNode, constraint);

        TGConnectingPoint p1 = constraint.closerFreeTGConnectingPoint(attack.getX(), attack.getY(), false, true);
        TGConnectingPoint p2 = attack.closerFreeTGConnectingPoint(constraint.getX(), constraint.getY(), true, false);

        if (p1 != null && p2 != null) {
            p1.setFree(false);
            p2.setFree(false);

            TGConnector conn = null;
            Vector<Point> points = new Vector<Point>();
            conn = new ATDAttackConnector(0, 0, 0, 0, 0, 0, true, null,
                    this, p1, p2, points);
            if (_valueConst != -1) {
                conn.setValue("" + _valueConst);
            }

            if (conn != null) {
                addComponent(conn, 0, 0, false, true);
            }
        }

        // We now create all the sub attacks
        int size = nextNode.getInputAttacks().size();
        int currentBeg = _x - _length/2;
        int slotSize = _length / size;
        for(int i=0; i<size ; i++) {
            Attack nextAttack = nextNode.getInputAttacks().get(i);
            Integer nextValue = nextNode.getInputValues().get(i);

            String s = drawAttackFromAttackTreeModel(_mapOfComponents, _at, nextAttack, constraint, nextValue, currentBeg+ slotSize/2 , _y+70,
                    slotSize);
            currentBeg += slotSize;

            if (s != null) {
                return s;
            }

        }


        return null;
    }
    
}
