package ai;

import myutil.TraceManager;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;

/**
 * Class AIUseCaseDiagram
 * <p>
 * Creation: 18/03/2024
 *
 * @author Alan Birchler De Allende
 */
public class CAPECTracer extends AIInteract {
    public static String QUESTIONTRACECAPECS = "From the provided system specifications, identify all of the possible " +
            "attack patterns that an attacker could use to exploit the system.";

    public CAPECTracer(AIChatData _chatData) {
        super(_chatData);
    }

    @Override
    public void internalRequest() {
        chatData.feedback.addToChat(QUESTIONTRACECAPECS, true);

        String systemSpec = chatData.lastQuestion.trim();

        Path capecTracerFolder = Paths.get(Objects.requireNonNull(
                this.getClass().getResource("../capec_tracer")).getPath());

        writeToFile(capecTracerFolder + "/system_specs.txt", systemSpec);

        String results = runCapecTracer(capecTracerFolder.toString());
        chatData.feedback.addToChat(results, false);
    }

    private void writeToFile(String filePath, String content) {
        try {
            // Create a FileWriter object
            FileWriter fileWriter = new FileWriter(filePath, false);

            // Wrap the FileWriter in a BufferedWriter for efficient writing
            BufferedWriter bufferedWriter = new BufferedWriter(fileWriter);

            // Write the content to the file
            bufferedWriter.write(content);

            // Close the BufferedWriter to flush the buffer and close the file
            bufferedWriter.close();

        } catch (IOException e) {
            TraceManager.addDev(e.getMessage());
        }
    }

    private String runCapecTracer(String capecTracerFolder) {
        String traces = "";

        try {
            String s = null;

            // Execute the process
            Process p = Runtime.getRuntime().exec("python3 " + capecTracerFolder + "/capec_tracer.py");

            BufferedReader stdInput = new BufferedReader(new
                    InputStreamReader(p.getInputStream()));
            BufferedReader stdError = new BufferedReader(new
                    InputStreamReader(p.getErrorStream()));

            // read the output from the command
            while ((s = stdInput.readLine()) != null) {
                TraceManager.addDev(s);
            }

            // read any errors from the attempted command
            while ((s = stdError.readLine()) != null) {
                TraceManager.addDev(s);
            }

            byte[] bytes = Files.readAllBytes(Path.of(capecTracerFolder + "/traced_capecs.txt"));

            traces = new String(bytes);
        } catch (IOException e) {
            TraceManager.addDev(e.getMessage());
        }

        if (traces.isEmpty()) {
            return "The tracer failed to run successfully.";
        }
        else {
            return traces;
        }
    }

    @Override
    public Object applyAnswer(Object input) {
        return null;
    }
}
